# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import socket
import logging
from ConfigParser import ConfigParser
from os.path import exists
from gettext import gettext as _

import dbus
import gobject
import gconf

from sugar_client import util, env, service
from sugar_client.rest import JSONRest
from sugar_client.util import enforce


INTERFACE = 'org.sugarlabs.client.Server'
NAME = 'org.sugarlabs.client.Server'
PATH = '/org/sugarlabs/client/Server'

_XO_SERIAL_PATH = '/ofw/mfg-data/SN'
_XO_UUID_PATH = '/ofw/mfg-data/U#'
_NICKNAME_GCONF = '/desktop/sugar/user/nick'


class Service(service.Service):

    def __init__(self):
        service.Service.__init__(self, dbus.SessionBus, INTERFACE, NAME, PATH)

    @env.BaseService.method(INTERFACE,
            in_signature='s', out_signature='b')
    def Register(self, api_url):
        if not self.acquire():
            return False

        def register():
            try:
                self._register(api_url or env.api_url.value)
            except Exception, error:
                self.info(_('Cannot process registration: %s'), error)
                util.exception(_('Cannot process registration'))
                self.finished(service.STATE_FAILED)
            else:
                self.info(_('Registration processed successfully'))
                self.finished(service.STATE_SUCCESS)

        gobject.idle_add(register)
        return True

    @property
    def Registered(self):
        return (self.update_status() or {}).get('registered', False)

    @env.BaseService.method(INTERFACE,
            in_signature='', out_signature='')
    def Cancel(self):
        raise NotImplementedError()

    def update_status(self):
        """Fetch status from a server and update local config correspondingly.

        :returns:
            on successful call, returns the status dictionary and None othewise

        """
        if not env.uid.value:
            return

        try:
            rest = JSONRest(env.api_url.value)
            reply = rest.get('/client/status', {'uid': env.uid.value})

            if reply.get('registered'):
                config = {}
                for i in ['jabber-url', 'backup-url']:
                    if i in reply:
                        config[i] = reply[i]
                if 'jabber-url' in config:
                    self._set_jabber(config['jabber-url'])
                config['schoolserver'] = socket.getfqdn('schoolserver')
                self._update_profile_config(config)

            return reply
        except Exception:
            util.exception(_('Cannot connect to school server %s ' \
                    'to fetch client status'), env.api_url.value)

    def _register(self, api_url):
        self.info(_('Registring on %s'), api_url)

        pubkey_path = env.profile_path('owner.key.pub')
        enforce(exists(pubkey_path),
                _('Sugar session was never started, no pubkey'))

        request = {
                'machine_sn': env.machine_sn.value,
                'machine_uuid': env.machine_uuid.value,
                'nickname': env.nickname.value,
                'pubkey': self._read_pubkey(pubkey_path),
                }

        if request['machine_sn'] is None and exists(_XO_SERIAL_PATH):
            request['machine_sn'] = _read_XO_value(_XO_SERIAL_PATH)
        if request['machine_uuid'] is None and exists(_XO_UUID_PATH):
            request['machine_uuid'] = _read_XO_value(_XO_UUID_PATH)
        if request['nickname'] is None:
            gconf_client = gconf.client_get_default()
            request['nickname'] = gconf_client.get_string(_NICKNAME_GCONF)

        reply = JSONRest(api_url).post('/client/register', request)

        logging.debug('Got registration reply: %r', reply)
        enforce(reply.get('success') == 'OK',
                _('Registration on %s was failed: %s'),
                api_url, reply.get('error') or _('Unknown error'))

        self._update_profile_config({
            'api-url': api_url,
            'uid': reply['uid'],
            'schoolserver': socket.getfqdn('schoolserver'),
            })

    def _set_jabber(self, url):
        self.info(_('Setting Jabber server to %s'), url)

        # Calling sugar-control-panel will trigger TP reconnection
        if util.call(['sugar-control-panel', '-s', 'jabber', url]) is not None:
            return

        logging.warn(_('sugar-control-panel failed, ' \
                'will set jabber directly in GConf'))

        gconf_client = gconf.client_get_default()
        gconf_client.set_string('/desktop/sugar/collaboration/jabber_server',
                url)

    def _read_pubkey(self, path):
        for line in file(path).readlines():
            line = line.strip()
            if line.startswith('ssh-'):
                return line
        raise RuntimeError(_('Valid SSH public key was not found in %s') % \
                path)

    def _update_profile_config(self, data):
        config_path = env.profile_path('sugar-client.conf')

        config = ConfigParser()
        if exists(config_path):
            config.read(config_path)
        if not config.has_section('main'):
            config.add_section('main')

        for key, value in data.items():
            config.set('main', key, value)
            if key in util.Option.items:
                util.Option.items[key].value = value

        config_file = file(config_path, 'w')
        config.write(config_file)
        config_file.close()


def _read_XO_value(path):
    return file(path).read().rstrip('\0\n')
